/*
*atmel9263fb.c--Low level implementation of the 
*Atmel 9263 RT12232C LCD frame buffer device 
*
*(c)Copyright Honeywell 2006-2010 Cw.feng 
*
*this version is based on :
*
*linux/drivers/video/vfb.c--Virtual frame buffer device 
*
*This file is subject to the terms and conditions of the GNU General Public 
*License.See the file COPYING in the main directory of this archive for 
*more details.
*
*Layout is based on skeletonfb.c by James Simmons and Geert Uytterhoeven.
*/
#include<linux/module.h>
#include<linux/kernel.h>
#include<linux/errno.h>
#include<linux/string.h>
#include<linux/mm.h>
#include<linux/tty.h>
#include<linux/slab.h>
#include<linux/vmalloc.h>
#include<linux/delay.h>
#include<linux/fb.h>
#include<linux/init.h>
#include<linux/time.h>
#include<linux/platform_device.h>
#include<linux/netcorefb.h>
#include "AT91SAM9263_inc.h" 
#include<asm/uaccess.h>
#include<asm/io.h>

#define BANKWIDTH 61 

#define floor8(a)(a&(~0x07))
#define iceil8(a)(((int)((a+7)/8))*8)
#define ceilX(a,x)(((a+x)/x)*x-1)

/* Atmel RT12232 chipset specific defines and code */
#define NC_SET_DISPLAY_START_LINE 0x00C0 
#define NC_SET_PAGE_ADDR 0x00B8 
#define NC_SET_COLUMN 0x0000 
#define NC_SEL_ADC_CW 0x00A0 
#define NC_SEL_ADC_CCW 0x00A1 
#define NC_STATIC_DRV_ON 0x00A5 
#define NC_STATIC_DRV_OFF 0x00A4 
#define NC_RD_MOD_WR 0x00E0 
#define NC_END 0x00EE 
#define NC_RESET 0x00E2 
#define NC_DISPLAY_ON 0x00AF 
#define NC_DISPLAY_OFF 0x00AE 
#define NC_BACK_LIGHT_ON 0x0002 
#define NC_BACK_LIGHT_OFF 0x0000 
#define NC_RESET_PIN_DOWN 0x0000 
#define NC_RESET_PIN_HIGH 0x0001 
#define NC_SEL_DUTY_32 0x00A9 
#define NC_SEL_DUTY_16 0x00A8 

/* LCD status mask */
#define NC_LCD_STATUS_RESTE 0x0010 
#define NC_LCD_STATUS_DISPLAY 0x0020 
#define NC_LCD_STATUS_ADC 0x0040 
#define NC_LCD_STATUS_BUSY 0x0080 

#define NUM_OF_BANKS 2 
#define NUM_OF_PAGES_PER_BANK 4 
#define NUM_OF_COLUMNS_PER_BANK 60 

static unsigned char atmel9263_page[122];
struct atmel9263fb_par 
{
    unsigned long dio_addr[NUM_OF_BANKS];unsigned long cio_addr[NUM_OF_BANKS];
    unsigned long cmain_addr ;
    atomic_t ref_count ;
    struct fb_info*info ;
    spinlock_t lock ;
}
;

static struct fb_fix_screeninfo atmel9263fb_fix __initdata=
{
    .id="atmel9263fb",
    .type=FB_TYPE_PACKED_PIXELS,
    .visual=FB_VISUAL_MONO01,
    .xpanstep=0,
    .ypanstep=1,
    .ywrapstep=0,
    .accel=FB_ACCEL_NONE,
}
;

static struct fb_var_screeninfo atmel9263fb_var __initdata=
{
    .xres=120,
    .yres=32,
    .xres_virtual=120,
    .yres_virtual=32,
    .bits_per_pixel=1,
    .nonstd=1,
}
;

static struct platform_device*atmel9263fb_device ;

static unsigned long cmain_addr ;
static unsigned long dio_addr[NUM_OF_BANKS];
static unsigned long cio_addr[NUM_OF_BANKS];

//#define debug 
//#define time_reset
#define uboot_init

#define NWE_SETUP(time)(time<<0)
#define NCS_SETUP_WR(time)(time<<8)
#define NRD_SETUP(time)(time<<16)
#define NCS_SETUP_RD(time)(time<<24)

#define NWE_PULSE(time)(time<<0)
#define NCS_PULSE_WR(time)(time<<8)
#define NRD_PULSE(time)(time<<16)
#define NCS_PULSE_RD(time)(time<<24)

#define NWE_CYCLE(time)(time<<0)
#define NRD_CYCLE(time)(time<<16)

static unsigned char*PIOE_Base_Address,*PIOB_Base_Address ;
static unsigned char*EBI1_CS0_Addr,*EBI1_CS2_Addr ;
#ifdef time_reset 
#define DELAY_TIME 500 
static void lcd_test(void);
DEFINE_TIMER(reset_timer,lcd_test,0,NULL);

static const unsigned char UltraKeyLite_Init_left[]=
{
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x98,0x98,0x00,0x00,0x00,0x00,0x00,0x00,0x98,0x98,
    0x00,0x00,0x00,0x00,0x00,0x00,0x98,0x98,0x00,0x00,0x00,0x00,0x00,0x04,0xFC,0xFC,0x04,0x00,0x00,0x00,0x00,
    0xF0,0xF0,0x20,0x30,0xF0,0xE0,0x00,0x00,0x00,0x00,0x20,0xEC,0xEC,0x00,0x00,0x00,0x00,0x60,0xF8,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x19,0x19,0x00,0x00,0x00,0x00,0x00,0x00,0x19,0x19,0x00,0x00,
    0x00,0x00,0x00,0x00,0x19,0x19,0x00,0x00,0x00,0x00,0x00,0x20,0x3F,0x3F,0x20,0x00,0x00,0x00,0x00,0x3F,0x3F,
    0x00,0x00,0x3F,0x3F,0x00,0x00,0x00,0x00,0x20,0x3F,0x3F,0x20,0x00,0x00,0x00,0x00,0x3F,
}
;
static const unsigned char UltraKeyLite_Init_right[]=
{
    0xF8,0x60,0x60,0x00,0x00,0x00,0x20,0xEC,0xEC,0x00,0x00,0x00,0x00,0x00,0xE0,0xF0,0x30,0x30,0xF0,0xE0,0x00,
    0x00,0x00,0x00,0x00,0xFC,0xFC,0x00,0x00,0x00,0x00,0x00,0x98,0x98,0x00,0x00,0x00,0x00,0x00,0x00,0x98,0x98,
    0x00,0x00,0x00,0x00,0x00,0x00,0x98,0x98,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x3F,0x30,
    0x30,0x00,0x00,0x00,0x20,0x3F,0x3F,0x20,0x00,0x00,0x00,0x00,0x1E,0x3E,0x33,0x11,0x3F,0x3F,0x00,0x00,0x00,
    0x00,0x00,0x3F,0x3F,0x00,0x00,0x00,0x00,0x00,0x19,0x19,0x00,0x00,0x00,0x00,0x00,0x00,0x19,0x19,0x00,0x00,
    0x00,0x00,0x00,0x00,0x19,0x19,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
}
;
static const unsigned char UltraKeyLite_Waiting_right[]=
{
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x3C,0xFC,0xE0,0x00,0xF8,0x7C,0xF8,0x00,0xE0,0xFC,0x3C,
    0x00,0xE0,0xF0,0x30,0x30,0xF0,0xE0,0x00,0x00,0x20,0xEC,0xEC,0x00,0x00,0x60,0xF8,0xF8,0x60,0x60,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x1F,0x3F,0x30,0x1F,0x00,0x1F,0x30,0x3F,0x1F,0x00,0x00,0x1E,
    0x3E,0x33,0x11,0x3F,0x3F,0x00,0x00,0x20,0x3F,0x3F,0x20,0x00,0x00,0x3F,0x3F,0x30,0x30,
}
;
static const unsigned char UltraKeyLite_Waiting_left[]=
{
    0x20,0xEC,0xEC,0x00,0x00,0xF0,0xF0,0x20,0x30,0xF0,0xE0,0x00,0x00,0xF0,0xF0,0x30,0x20,0xF0,0xF0,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x20,0x3F,0x3F,
    0x20,0x00,0x3F,0x3F,0x00,0x00,0x3F,0x3F,0x00,0x00,0x33,0x77,0x66,0x62,0x7F,0x3F,0x00,0x00,0x00,0x03,0x03,
    0x00,0x03,0x03,0x00,0x03,0x03,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
}
;
#endif 

static inline void AT9263_lcd_init(void)
{
    unsigned long ret ;
    unsigned char*EBI1CSA_Addr,*SMC_Addr ;
    unsigned char i,j,k ;
    
    #ifdef uboot_init
    PIOE_Base_Address=ioremap(AT91C_BASE_PIOE,0xAC);
    // configure PIOE22 to peripheral B with pull up
    iowrite32(AT91C_PE22_EBI1_NCS2_NANDCS,PIOE_Base_Address+PIO_PDR(0));    
    iowrite32(AT91C_PE22_EBI1_NCS2_NANDCS,PIOE_Base_Address+PIO_ODR(0));
    iowrite32(AT91C_PE22_EBI1_NCS2_NANDCS,PIOE_Base_Address+PIO_IFDR(0));
    iowrite32(AT91C_PE22_EBI1_NCS2_NANDCS,PIOE_Base_Address+PIO_CODR(0));
    iowrite32(AT91C_PE22_EBI1_NCS2_NANDCS,PIOE_Base_Address+PIO_IDR(0));
    iowrite32(AT91C_PE22_EBI1_NCS2_NANDCS,PIOE_Base_Address+PIO_MDDR(0));
    iowrite32(AT91C_PE22_EBI1_NCS2_NANDCS,PIOE_Base_Address+PIO_PPUSR(0));
    iowrite32(AT91C_PE22_EBI1_NCS2_NANDCS,PIOE_Base_Address+PIO_BSR(0));
    iowrite32(AT91C_PE22_EBI1_NCS2_NANDCS,PIOE_Base_Address+PIO_OWDR(0));

    #ifdef debug 
    printk("PIOE_Base_Address = 0x%x\n",PIOE_Base_Address);
    printk("AT91C_PIOE_PER = 0x%x\n",ioread32(PIOE_Base_Address+PIO_PSR(0)));
    printk("AT91C_PIOE_PDR = 0x%x\n",~(ioread32(PIOE_Base_Address+PIO_PSR(0))));
    printk("AT91C_PIOE_OER = 0x%x\n",ioread32(PIOE_Base_Address+PIO_OSR(0)));
    printk("AT91C_PIOE_ODR = 0x%x\n",~(ioread32(PIOE_Base_Address+PIO_OSR(0))));
    printk("AT91C_PIOE_IFER = 0x%x\n",ioread32(PIOE_Base_Address+PIO_IFSR(0)));
    printk("AT91C_PIOE_IFDR = 0x%x\n",~(ioread32(PIOE_Base_Address+PIO_IFSR(0))));
    printk("AT91C_PIOE_SODR = 0x%x\n",ioread32(PIOE_Base_Address+PIO_ODSR(0)));
    printk("AT91C_PIOE_CODR = 0x%x\n",~(ioread32(PIOE_Base_Address+PIO_ODSR(0))));
    printk("AT91C_PIOE_IER = 0x%x\n",ioread32(PIOE_Base_Address+PIO_IMR(0)));
    printk("AT91C_PIOE_IDR = 0x%x\n",~(ioread32(PIOE_Base_Address+PIO_IMR(0))));
    printk("AT91C_PIOE_MDER = 0x%x\n",ioread32(PIOE_Base_Address+PIO_MDSR(0)));
    printk("AT91C_PIOE_MDDR = 0x%x\n",~(ioread32(PIOE_Base_Address+PIO_MDSR(0))));
    printk("AT91C_PIOE_PPUDR = 0x%x\n",ioread32(PIOE_Base_Address+PIO_PPUSR(0)));
    printk("AT91C_PIOE_PPUER = 0x%x\n",~(ioread32(PIOE_Base_Address+PIO_PPUSR(0))));
    printk("AT91C_PIOE_ASR = 0x%x\n",~(ioread32(PIOE_Base_Address+PIO_ABSR(0))));
    printk("AT91C_PIOE_BSR = 0x%x\n",ioread32(PIOE_Base_Address+PIO_ABSR(0)));
    printk("AT91C_PIOE_OWER = 0x%x\n",ioread32(PIOE_Base_Address+PIO_OWSR(0)));
    printk("AT91C_PIOE_OWDR = 0x%x\n",~(ioread32(PIOE_Base_Address+PIO_OWSR(0))));
    #endif 
    iounmap(PIOE_Base_Address);
    
    //configure EBI1 CSA reg of CS2 to  NCS2
    EBI1CSA_Addr=ioremap(AT91C_CCFG_EBI1CSA,4);
    ret=ioread32(EBI1CSA_Addr);
    iowrite32(ret&~AT91C_EBI_CS2A_SM,EBI1CSA_Addr);
    //ret = (unsigned int)inw(AT91C_CCFG_EBI1CSA);
    //outw(ret & ~AT91C_EBI_CS2A_SM, AT91C_CCFG_EBI1CSA);
    #ifdef debug 
    printk("AT91C_CCFG_EBI1CSA = 0x%x\n",ioread32(EBI1CSA_Addr));
    #endif 
    iounmap(EBI1CSA_Addr);
    //configure SMC MODE reg
    SMC_Addr=ioremap(AT91C_BASE_SMC1,16*3);
    iowrite32(NWE_SETUP(26)|NCS_SETUP_WR(31)|NRD_SETUP(26)|NCS_SETUP_RD(31),SMC_Addr+SMC_SETUP0);
    iowrite32(NWE_PULSE(58)|NCS_PULSE_WR(50)|NRD_PULSE(58)|NCS_PULSE_RD(50),SMC_Addr+SMC_PULSE0);
    iowrite32(NWE_CYCLE(90)|NRD_CYCLE(90),SMC_Addr+SMC_CYCLE0);
    iowrite32(0x0,SMC_Addr+SMC_CTRL0);
    
    iowrite32(NWE_SETUP(26)|NCS_SETUP_WR(31)|NRD_SETUP(26)|NCS_SETUP_RD(31),SMC_Addr+SMC_SETUP2);
    iowrite32(NWE_PULSE(58)|NCS_PULSE_WR(50)|NRD_PULSE(58)|NCS_PULSE_RD(50),SMC_Addr+SMC_PULSE2);
    iowrite32(NWE_CYCLE(90)|NRD_CYCLE(90),SMC_Addr+SMC_CYCLE2);
    iowrite32(0x0,SMC_Addr+SMC_CTRL2);
    #ifdef debug 
    printk("AT91C_SMC1_SETUP0 = 0x%x\n",ioread32(SMC_Addr+SMC_SETUP0));
    printk("AT91C_SMC1_PULSE0 = 0x%x\n",ioread32(SMC_Addr+SMC_PULSE0));
    printk("AT91C_SMC1_CYCLE0 = 0x%x\n",ioread32(SMC_Addr+SMC_CYCLE0));
    printk("AT91C_SMC1_CTRL0 = 0x%x\n",ioread32(SMC_Addr+SMC_CTRL0));
    
    printk("AT91C_SMC1_SETUP2 = 0x%x\n",ioread32(SMC_Addr+SMC_SETUP2));
    printk("AT91C_SMC1_PULSE2 = 0x%x\n",ioread32(SMC_Addr+SMC_PULSE2));
    printk("AT91C_SMC1_CYCLE2 = 0x%x\n",ioread32(SMC_Addr+SMC_CYCLE2));
    printk("AT91C_SMC1_CTRL2 = 0x%x\n",ioread32(SMC_Addr+SMC_CTRL2));
    #endif 
    iounmap(SMC_Addr);
#endif 
    //configure data address
    EBI1_CS0_Addr=ioremap(AT91C_EBI1_CS0,2);
    EBI1_CS2_Addr=ioremap(AT91C_EBI1_CS2,2);
    
    cio_addr[0]=(unsigned long)EBI1_CS0_Addr ;    
    cio_addr[1]=(unsigned long)EBI1_CS2_Addr ;
    
    dio_addr[0]=(unsigned long)(EBI1_CS0_Addr+1);
    dio_addr[1]=(unsigned long)(EBI1_CS2_Addr+1);
    
#ifdef uboot_init
    #ifdef debug 
    printk("EBI1_CS0_Addr = 0x%x EBI1_CS2_Addr = 0x%x\n",EBI1_CS0_Addr,EBI1_CS2_Addr);
    #endif 
    for (i=0;i<2;i++)
    {
        iowrite8(NC_RESET,cio_addr[i]);
        iowrite8(NC_DISPLAY_OFF,cio_addr[i]);
        iowrite8(NC_STATIC_DRV_OFF,cio_addr[i]);
        iowrite8(NC_SEL_DUTY_32,cio_addr[i]);
        iowrite8(NC_SEL_ADC_CW,cio_addr[i]);
        iowrite8(NC_END,cio_addr[i]);
        iowrite8(NC_SET_COLUMN,cio_addr[i]);
        iowrite8(NC_SET_DISPLAY_START_LINE,cio_addr[i]);
        iowrite8(NC_DISPLAY_ON,cio_addr[i]);
/*        
        for(j=0;j<4;j++)
        {
            iowrite8(((j&0x0003)|NC_SET_PAGE_ADDR),cio_addr[i]);
            for(k=0;k<80;k++)
            {
                iowrite8((k|NC_SET_COLUMN),cio_addr[i]);
                iowrite8(0x0,dio_addr[i]);
            }
        }
 */       
        iowrite8(NC_SET_PAGE_ADDR,cio_addr[i]);
        iowrite8(NC_SET_COLUMN,cio_addr[i]);
        
#ifdef debug 
        printk("LCD status = 0x%x\n",ioread8(cio_addr[i]));
        printk("LCD data   = 0x%x\n",ioread8(dio_addr[i]));
#endif 
    }
#endif            
    PIOB_Base_Address=ioremap(AT91C_BASE_PIOB,0xAC);

#ifdef uboot_init
    //configure PIOB8 
    iowrite32(AT91C_PIO_PB8 |AT91C_PIO_PB9,(PIOB_Base_Address+PIO_IDR(0)));
    iowrite32(AT91C_PIO_PB8 |AT91C_PIO_PB9,(PIOB_Base_Address+PIO_PPUDR(0)));
    iowrite32(AT91C_PIO_PB8 |AT91C_PIO_PB9,(PIOB_Base_Address+PIO_SODR(0)));    // enable lcd backlight
    iowrite32(AT91C_PIO_PB8 |AT91C_PIO_PB9,(PIOB_Base_Address+PIO_PER(0)));
    iowrite32(AT91C_PIO_PB8 |AT91C_PIO_PB9,(PIOB_Base_Address+PIO_OER(0)));
    
    #ifdef debug 
    printk("PIOE_Base_Address = 0x%x\n",PIOB_Base_Address);
    printk("AT91C_PIOB_IER = 0x%x\n",ioread32(PIOB_Base_Address+PIO_IMR(0)));
    printk("AT91C_PIOB_IDR = 0x%x\n",~(ioread32(PIOB_Base_Address+PIO_IMR(0))));
    printk("AT91C_PIOB_PPUER = 0x%x\n",ioread32(PIOB_Base_Address+PIO_PPUSR(0)));
    printk("AT91C_PIOB_PPUDR = 0x%x\n",~(ioread32(PIOB_Base_Address+PIO_PPUSR(0))));
    printk("AT91C_PIOB_SODR = 0x%x\n",ioread32(PIOB_Base_Address+PIO_PDSR(0)));
    printk("AT91C_PIOB_CODR = 0x%x\n",~(ioread32(PIOB_Base_Address+PIO_PDSR(0))));
    printk("AT91C_PIOB_PER = 0x%x\n",ioread32(PIOB_Base_Address+PIO_PSR(0)));
    printk("AT91C_PIOB_PDR = 0x%x\n",~(ioread32(PIOB_Base_Address+PIO_PSR(0))));
    printk("AT91C_PIOB_OER = 0x%x\n",ioread32(PIOB_Base_Address+PIO_OSR(0)));
    printk("AT91C_PIOB_ODR = 0x%x\n",~(ioread32(PIOB_Base_Address+PIO_OSR(0))));
    #endif 
#endif    
    cmain_addr=(unsigned long)PIOB_Base_Address ;
        
}

#ifdef time_reset 
static void lcd_test(void)
{
    unsigned char i,j ;
    static unsigned char num=0 ;
    
    if(num==0)
    {
        for(i=0;i<2;i++)
        {
            iowrite8(((i+2)&0x0003)|NC_SET_PAGE_ADDR,cio_addr[0]);
            for(j=0;j<61;j++)
            {
                iowrite8((j|NC_SET_COLUMN),cio_addr[0]);
                //Reset column to 0            
                iowrite8(UltraKeyLite_Init_left[j+i*61],dio_addr[0]);
            }
        }
        for(i=0;i<2;i++)
        {
            iowrite8(((i+2)&0x0003)|NC_SET_PAGE_ADDR,cio_addr[1]);
            for(j=0;j<61;j++)
            {
                iowrite8((j|NC_SET_COLUMN),cio_addr[1]);
                //Reset column to 0                 
                iowrite8(UltraKeyLite_Init_right[j+i*61],dio_addr[1]);
            }
        }
        for(i=0;i<2;i++)
        {
            iowrite8((i&0x0003)|NC_SET_PAGE_ADDR,cio_addr[0]);
            for(j=0;j<61;j++)
            {
                iowrite8((j|NC_SET_COLUMN),cio_addr[0]);
                //Reset column to 0            
                iowrite8(UltraKeyLite_Waiting_right[j+i*61],dio_addr[0]);
            }
        }
        for(i=0;i<2;i++)
        {
            iowrite8((i&0x0003)|NC_SET_PAGE_ADDR,cio_addr[1]);
            for(j=0;j<60;j++)
            {
                iowrite8((j|NC_SET_COLUMN),cio_addr[1]);
                //Reset column to 0                 
                iowrite8(UltraKeyLite_Waiting_left[j+i*60],dio_addr[1]);
            }
        }
        num =1;
    }
#ifdef debug
    printk("LCD status = 0x%x\n",ioread8(cio_addr[0]));
    printk("LCD data   = 0x%x\n",ioread8(dio_addr[0]));
    printk("LCD status = 0x%x\n",ioread8(cio_addr[1]));
    printk("LCD data   = 0x%x\n",ioread8(dio_addr[1]));
#endif
    //reset_timer.expires+=DELAY_TIME ;
    //add_timer(&reset_timer);
}
#endif 
static inline void lcd_write_ctl(struct atmel9263fb_par*par,unsigned int value)
{
    iowrite32(AT91C_PIO_PB8,value?((par->cmain_addr)+PIO_SODR(0)):((par->cmain_addr)+PIO_CODR(0)));
}


static inline void lcd_bank_write_data(struct atmel9263fb_par*par,unsigned int bank,
unsigned char value)
{
    iowrite8(value,par->dio_addr[bank]);
}

static inline unsigned char lcd_bank_read_data(struct atmel9263fb_par*par,unsigned int bank)
{
    return ioread8(par->dio_addr[bank]);
}


static inline void lcd_bank_write_ctl(struct atmel9263fb_par*par,unsigned int bank,
unsigned short value)
{
    iowrite8(value,par->cio_addr[bank]);
}


static inline unsigned char lcd_bank_read_ctl(struct atmel9263fb_par*par,unsigned int bank)
{
    return ioread8(par->cio_addr[bank]);
    
}


/* main atmel9263fb operations */
static int atmel9263fb_open(struct fb_info*info,int user)
{
    struct atmel9263fb_par*par=info->par ;
    
    atomic_inc(&par->ref_count);
    return 0 ;
}

static int atmel9263fb_release(struct fb_info*info,int user)
{
    struct atmel9263fb_par*par=info->par ;
    int count=atomic_read(&par->ref_count);
    
    if(!count)
    return-EINVAL ;
    atomic_dec(&par->ref_count);
    return 0 ;
}

static int atmel9263fb_pan_display(struct fb_var_screeninfo*var,
struct fb_info*info)
{
    int i ;
    struct atmel9263fb_par*par=info->par ;
    
    if((var->vmode&FB_VMODE_YWRAP)&&(var->yoffset<32))
    {
        for(i=0;i<NUM_OF_BANKS;i++)
        {
            /* Set new startline */
            lcd_bank_write_ctl(par,i,((var->yoffset)&0x001F)|NC_SET_DISPLAY_START_LINE);
        }
        info->var.yoffset=var->yoffset ;
        return 0 ;
    }
    
    return-EINVAL ;
}

/*
 * here we handle a specific page on the lcd.
 */
static void atmel9263fb_lcd_update_vert(struct atmel9263fb_par*par,unsigned int upper,unsigned int left,unsigned int right,unsigned int bank)
{
    unsigned int src ;
    unsigned int linesize ;
    int i,j ;
    unsigned char val,tmp ;
    unsigned char bitmask ;
    
    
    /* Set page address */
    if(upper<2)
    {
        lcd_bank_write_ctl(par,bank,((upper/8+2)&0x0003)|NC_SET_PAGE_ADDR);
    }
    else 
    {
        lcd_bank_write_ctl(par,bank,((upper/8-2)&0x0003)|NC_SET_PAGE_ADDR);
    }
    
    /* Start byte in frame buffer */
    linesize=par->info->var.xres ;
    
    /* Set column address Column 61 is not used*/
    if(left>60)
    lcd_bank_write_ctl(par,1,((left-61)&0x007F)|NC_SET_COLUMN);
    else 
    lcd_bank_write_ctl(par,0,(left&0x007F)|NC_SET_COLUMN);
    
    bitmask=0x80 ;
    j=0 ;
    
    while(left<=right)
    {
        val=0 ;
        src=left+upper*linesize ;
        for(i=0;i<8;i++)
        {
            tmp=(*(par->info->screen_base+(src+i*linesize)/8))&(bitmask>>((src+i*linesize)%8));
            tmp>>=(8-1-(src+i*linesize)%8);
            val|=tmp<<i ;
            atmel9263_page[j]=val ;
        }
        /* Write one byte data to lcd */
        //lcd_bank_write_data(par, bank, val);
        //lcd_bank_write_data(par, bank, 0xff);
        left++;
        j++;
    }
    for(i=0;i<j;i++)
    {
        lcd_bank_write_data(par,bank,atmel9263_page[i]);
    }
}

/*
 * here we handle horizontal blocks for the update. update_vert will
 * handle spaning multiple pages.
 */
static void atmel9263fb_lcd_update_horiz(struct atmel9263fb_par*par,unsigned int left,
unsigned int right,unsigned int top,unsigned int h,unsigned int bank)
{
    unsigned int distance,upper,lower ;
    distance=h ;
    upper=floor8(top);
    lower=min(upper+distance-1,ceilX(upper,8));
    
    while(distance>0)
    {
        distance-=((lower-upper)+1);
        atmel9263fb_lcd_update_vert(par,upper,left,right,bank);
        upper=lower+1 ;
        lower=upper+7 ;
    }
}

/*
 * here we start the process of spliting out the fb update into
 * individual blocks of pixels.
 */
static void atmel9263fb_lcd_update(struct atmel9263fb_par*par,unsigned int dx,
unsigned int dy,unsigned int w,unsigned int h)
{
    unsigned int left,right,distance,y,bank=0 ;
    
    /* align the request first */
    y=floor8(dy);
    h+=dy-y ;
    h=iceil8(h);
    distance=w ;
    left=dx ;
    right=min(left+w-1,ceilX(left,BANKWIDTH));
    
    if(dx>60)
    {
        bank=1 ;
    }
    
    while(distance>0)
    {
        atmel9263fb_lcd_update_horiz(par,left,right,y,h,bank);
        distance-=((right-left)+1);
        left=right+1 ;
        right=min(left+distance-1,ceilX(left,BANKWIDTH));
        bank++;
    }
}

static void atmel9263fb_fillrect(struct fb_info*info,
const struct fb_fillrect*rect)
{
    struct atmel9263fb_par*par=info->par ;
    cfb_fillrect(info,rect);
    /* update the physical lcd */
    atmel9263fb_lcd_update(par,rect->dx,rect->dy,rect->width,rect->height);
}

static void atmel9263fb_copyarea(struct fb_info*info,
const struct fb_copyarea*area)
{
    struct atmel9263fb_par*par=info->par ;
    cfb_copyarea(info,area);
    /* update the physical lcd */
    atmel9263fb_lcd_update(par,area->dx,area->dy,area->width,area->height);
}

/* FIXME: This is a simple imageblit limited with no byte align image operate */
static void mcfb_imageblit(struct fb_info*info,struct fb_image*image)
{
    unsigned int dx,dy,w,h ;
    unsigned int startbyte,bytes_per_line ;
    int i,j ;
    unsigned char char_pre,char_next=0,tmp_char ;
    int offset ;
    
    dx=image->dx ;
    dy=image->dy ;
    w=image->width ;
    h=image->height ;
    
    
    if(!(dx%8))
    {
        startbyte=dx/8 ;
        bytes_per_line=w/8 ;
        for(i=0;i<h;i++)
        {
            for(j=0;j<bytes_per_line;j++)
            {
                get_user(tmp_char,(image->data+bytes_per_line*i+j));
                *(info->screen_base+15*(i+dy)+startbyte+j)=tmp_char ;
                #ifdef TIMER_RESET 
                *(videomemorybackup+15*(i+dy)+startbyte+j)=tmp_char ;
                #endif 
            }
        }
    }
    else 
    {
        startbyte=dx/8 ;
        offset=dx%8 ;
        bytes_per_line=w/8 ;
        for(i=0;i<h;i++)
        {
            for(j=0;j<bytes_per_line;j++)
            {
                get_user(tmp_char,(image->data+bytes_per_line*i+j));
                if(j==0)
                {
                    /* first byte */
                    //printk("startbyte is %d\n", 15*(i+dy) + startbyte);
                    char_pre=*(info->screen_base+15*(i+dy)+startbyte);
                    char_next=tmp_char ;
                    *((unsigned char*)(info->screen_base+15*(i+dy)+startbyte))=(unsigned char)((0xff<<(8-offset))&(char_pre))|(char_next>>offset);
                    #ifdef TIMER_RESET 
                    *((unsigned char*)(videomemorybackup+15*(i+dy)+startbyte))=(unsigned char)((0xff<<(8-offset))&(char_pre))|(char_next>>offset);
                    #endif 
                    //printk("value is 0x%x\n", *(info->screen_base + 15*(i+dy) + startbyte));
                }
                else 
                {
                    char_pre=char_next ;
                    char_next=tmp_char ;
                    *((unsigned char*)(info->screen_base+15*(i+dy)+startbyte+j))=(unsigned char)((char_pre)<<(8-offset))|(char_next>>(offset));
                    //printk("value is 0x%x\n", *(info->screen_base + 15*(i+dy) + startbyte + j));
                    #ifdef TIMER_RESET 
                    *((unsigned char*)(videomemorybackup+15*(i+dy)+startbyte+j))=(unsigned char)((char_pre)<<(8-offset))|(char_next>>(offset));
                    #endif 
                }
            }
            char_pre=char_next ;
            char_next=*(info->screen_base+15*(i+dy)+startbyte+bytes_per_line);
            *((unsigned char*)(info->screen_base+15*(i+dy)+startbyte+bytes_per_line))=(unsigned char)(char_pre<<(8-offset))|((0xff>>offset)&(char_next));
            #ifdef TIMER_RESET 
            *((unsigned char*)(videomemorybackup+15*(i+dy)+startbyte+bytes_per_line))=(unsigned char)(char_pre<<(8-offset))|((0xff>>offset)&(char_next));
            #endif 
            //printk("value is 0x%x\n", *(info->screen_base + 15*(i+dy) + startbyte + bytes_per_line));
        }
    }
    
}

static void atmel9263fb_imageblit(struct fb_info*info,const struct fb_image*image)
{
    struct atmel9263fb_par*par=info->par ;
    
    mcfb_imageblit(info,image);
    
    /* update the physical lcd */
    atmel9263fb_lcd_update(par,image->dx,image->dy,image->width,
    image->height);
}

static int atmel9263fb_ioctl(struct fb_info*info,unsigned int cmd,unsigned long arg)
{
    void __user*argp=(void __user*)arg ;
    struct atmel9263fb_par*par=info->par ;
    struct fb_image img ;
    unsigned int light_switch ;
    unsigned int ret ;
    
    switch(cmd)
    {
        case FBIO_IMAGEBLITE :
        {
            ret=copy_from_user(&img,argp,sizeof(img));
            info->fbops->fb_imageblit(info,&img);
            break ;
        }
        case FBIO_BACKLIGHTSWITCH :
        {
            ret=copy_from_user(&light_switch,argp,sizeof(unsigned int));
            lcd_write_ctl(par,light_switch);
            break ;
        }
        default :
        return-EINVAL ;
    }
    return 0 ;
}

/*
 * this is the access path from userspace. they can seek and write to
 * the fb. it's inefficient for them to do anything less than 60*8
 * writes since we update the lcd in each write() anyway.
 */
static int atmel9263fb_write(struct file*file,const char __user*buf,size_t count,
loff_t*ppos)
{
    struct inode*inode ;
    int fbidx ;
    struct fb_info*info ;
    unsigned long p ;
    int err=-EINVAL ;
    unsigned int fbmemlength,x,y,w,h,bitppos,startpos,endpos,bitcount ;
    struct atmel9263fb_par*par ;
    unsigned int xres ;
    
    p=*ppos ;
    inode=file->f_dentry->d_inode ;
    fbidx=iminor(inode);
    info=registered_fb[fbidx];
    if(!info||!info->screen_base)
    return-ENODEV ;
    
    par=info->par ;
    xres=info->var.xres ;
    fbmemlength=(xres*info->var.yres)/8 ;
    
    if(p>fbmemlength)
    return-ENOSPC ;
    
    err=0 ;
    if((count+p)>fbmemlength)
    {
        count=fbmemlength-p ;
        err=-ENOSPC ;
    }
    
    if(count)
    {
        char*base_addr ;
        
        base_addr=info->screen_base ;
        count-=copy_from_user(base_addr+p,buf,count);
        *ppos+=count ;
        err=-EFAULT ;
    }
    
    
    bitppos=p*8 ;
    startpos=bitppos/xres ;
    endpos=(bitppos+(count*8))%xres?(bitppos+(count*8))/xres+1:(bitppos+(count*8))/xres ;
    bitcount=endpos-startpos ;
    
    x=startpos%xres ;
    y=startpos/xres ;
    w=xres ;
    h=bitcount/xres ;
    atmel9263fb_lcd_update(par,x,y,w,h);
    
    if(count)
    return count ;
    return err ;
}

static struct fb_ops atmel9263fb_ops=
{
    .owner=THIS_MODULE,
    .fb_open=atmel9263fb_open,
    .fb_write=atmel9263fb_write,
    .fb_release=atmel9263fb_release,
    .fb_ioctl=atmel9263fb_ioctl,
    .fb_pan_display=atmel9263fb_pan_display,
    .fb_fillrect=atmel9263fb_fillrect,
    .fb_copyarea=atmel9263fb_copyarea,
    .fb_imageblit=atmel9263fb_imageblit,
}
;

static int __init atmel9263fb_probe(struct platform_device*dev)
{
    struct fb_info*info ;
    int retval=-ENOMEM ;
    int videomemorysize ;
    unsigned char*videomemory ;
    struct atmel9263fb_par*par ;
    unsigned char i ;
    
//fix the white dot ine the buttom of the LCD
    //videomemorysize=(120*32)/8 ;
	videomemorysize=(122*32)/8;    

    AT9263_lcd_init();
    if(!(videomemory=vmalloc(videomemorysize)))
    return retval ;
    
    memset(videomemory,0,videomemorysize);
    info=framebuffer_alloc(sizeof(struct atmel9263fb_par),&dev->dev);
    if(!info)
    goto err ;
    
    info->screen_base=(char __iomem*)videomemory ;
    info->fbops=&atmel9263fb_ops ;
    
    info->var=atmel9263fb_var ;
    info->fix=atmel9263fb_fix ;
    par=info->par ;
    par->info=info ;
    if(!dio_addr||!cio_addr||!cmain_addr)
    {
        printk(KERN_WARNING "no IO addresses supplied\n");
        goto err1 ;
    }
    
    par->cmain_addr=cmain_addr ;
    
    for(i=0;i<NUM_OF_BANKS;i++)
    {
        par->dio_addr[i]=dio_addr[i];
        par->cio_addr[i]=cio_addr[i];
    }
    info->flags=FBINFO_FLAG_DEFAULT ;
    spin_lock_init(&par->lock);
    retval=register_framebuffer(info);
    if(retval<0)
    goto err1 ;
    platform_set_drvdata(dev,info);
    printk(KERN_INFO 
    "fb%d: Atmel 9263 frame buffer device, using %d bytes of video memory\n",
    info->node,videomemorysize);
    #ifdef time_reset 
    reset_timer.data=(unsigned long)atmel9263fb_device ;
    reset_timer.function=lcd_test ;
    reset_timer.expires=jiffies+DELAY_TIME ;
    add_timer(&reset_timer);
    #endif 
    return 0 ;
    err1 :
    framebuffer_release(info);
    err :
    vfree(videomemory);
    return retval ;
}


static int atmel9263fb_remove(struct platform_device*dev)
{
    struct fb_info*info=platform_get_drvdata(dev);
    struct atmel9263fb_par*par=info->par ;
    
    if(info)
    {
        unregister_framebuffer(info);
        vfree(info->screen_base);
        #ifdef time_reset 
        del_timer_sync(&reset_timer);
        #endif 
        framebuffer_release(info);
    }
    lcd_write_ctl(par,NC_BACK_LIGHT_OFF);
    return 0 ;
}

static struct platform_driver atmel9263fb_driver=
{
    .probe=atmel9263fb_probe,
    .remove=atmel9263fb_remove,
    .driver=
    {
        .name="atmel9263fb",
    }
    ,
}
;

static int __init atmel9263fb_init(void)
{
    int ret ;
    ret=platform_driver_register(&atmel9263fb_driver);
    if(!ret)
    {
        atmel9263fb_device=platform_device_alloc("atmel9263fb",0);
        if(atmel9263fb_device)
        {
            ret=platform_device_add(atmel9263fb_device);
        }
        else 
        {
            ret=-ENOMEM ;
        }
        if(ret)
        {
            platform_device_put(atmel9263fb_device);
            platform_driver_unregister(&atmel9263fb_driver);
        }
    }
    return ret ;
}

static void __exit atmel9263fb_exit(void)
{
    platform_device_unregister(atmel9263fb_device);
    platform_driver_unregister(&atmel9263fb_driver);
}

module_init(atmel9263fb_init);
module_exit(atmel9263fb_exit);

MODULE_DESCRIPTION("fbdev driver for ATMEL 9263 LCD board");
MODULE_AUTHOR("Cw.feng");
MODULE_LICENSE("GPL");

